from openxm import OpenXMConst
import struct

class AbstractCMO:
    def __init__(self,head,xml_tag_name,body=None):
        self.head=head
        self.body=body
        self.xml_tag_name=xml_tag_name

    def get_cmo_tag(self):
        return self.head

    def head_to_byte(self,byte_order):
        return struct.pack(byte_order+'i',self.head)
    def body_to_byte(self,byte_order):
        if self.body!=None:
            return struct.pack(byte_order+'i',self.body)
        else:
            return ''
    def to_byte(self,byte_order):
        return self.head_to_byte(byte_order)+self.body_to_byte(byte_order)

    def to_python_object(self):
        return self.body


    def to_xml(self):
        return self.get_xml_open_tag()+self.get_xml_body_tag()+self.get_xml_close_tag()

    def get_xml_open_tag(self):
        return "<%s>" % self.xml_tag_name
        
    def get_xml_close_tag(self):
        return "</%s>" % self.xml_tag_name
        
    def get_xml_body_tag(self):
        if self.body != None:
            return "<int32>%d</int32>" % self.body
        else:
            return ""

class AbstractCMOWithCMO(AbstractCMO):
    def __init__(self,head,xml_tag_name,cmobj):
        AbstractCMO.__init__(self,head,xml_tag_name,cmobj)
    def body_to_byte(self,byte_order):
        return self.body.to_byte(byte_order)
    def get_xml_body_tag(self):
        return self.body.to_xml()
    def to_python_object(self):
        return (self.head ,self.body.to_python_object())
    @classmethod
    def new_from_stream(self,stream,byte_order):
        cmobj=cmo.new_from_stream(stream,byte_order)
        return self(cmobj)


class AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence(AbstractCMO):
    def __init__(self,head,xml_tag_name,info_fixedlength,seq_variablelength):
        AbstractCMO.__init__(self,head,xml_tag_name,(info_fixedlength,seq_variablelength))

    def get_len_of_sequence(self):
        return len(self.get_sequence())
    def get_sequence(self):
        (info,seq)=self.body
        return seq
    def get_info(self):
        (info,seq)=self.body
        return info

    def sequence_to_byte(self,byte_order):
        r=''
        for cmobj in self.get_sequence():
            r=r+cmobj.to_byte(byte_order)
        return r

    def info_to_byte(self,byte_order):
        info=self.get_info()
        if info==None:
            return ''
        r=''
        for cmobj in info:
            r=r+cmobj.to_byte(byte_order)
        return r
    def body_to_byte(self,byte_order):
        return struct.pack(byte_order+'i',self.get_len_of_sequence())+self.info_to_byte(byte_order)+self.sequence_to_byte(byte_order)

    def to_python_object(self):
        info=self.info_to_python_object()
        seq=self.sequence_to_python_object()
        if info==None:
            return seq
        else:
            return (info,seq)

    def sequence_to_python_object(self):
        return [elm.to_python_object() for elm in self.get_sequence()]
    def info_to_python_object(self):
        info=self.get_info()
        if info==None:
            return None
        else:
            return tuple([elm.to_python_object() for elm in info])



    def get_xml_sequence_tag(self):
        return "".join(i.to_xml() for i in self.get_sequence())
    def get_xml_info_tag(self):
        info=self.get_info()
        if info==None:
            return ''
        else:
            return "".join(i.to_xml() for i in info)
    def get_xml_body_tag(self):
        r='<int32 for="length">%d</int32>' % self.get_len_of_sequence()
        r=r+self.get_xml_info_tag()
        r=r+self.get_xml_sequence_tag()
        return r



class AbstractCMOWithSequence(AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence):
    def __init__(self,head,xml_tag_name,seq):
        AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence.__init__(self,head,xml_tag_name,None,seq)
    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        listofcmobj=[cmo.new_from_stream(stream,byte_order) for i in range(n)]
        return self(listofcmobj)


class AbstractCMOWithBytes(AbstractCMOWithSequence):
    def __init__(self,head,xml_tag_name,bytes):
        AbstractCMOWithSequence.__init__(self,head,xml_tag_name,bytes)
    def sequence_to_byte(self,byte_order):
        return self.get_sequence()
    def get_xml_sequence_tag(self):
        return "".join("<byte>%d</byte>" % i for i in self.get_sequence())
    def to_python_object(self):
        return self.get_sequence()

    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        byte_data=stream.read(n)
        return self(byte_data)


class AbstractCMOWithCMOTuple(AbstractCMO):
    def __init__(self,head,xml_tag_name,cmotuple):
        AbstractCMO.__init__(self,head,xml_tag_name,cmotuple)
    def body_to_byte(self,byte_order):
        r=""
        for cmoobj in self.body:
            r=r+cmoobj.to_byte(byte_order)
        return r
    def to_python_object(self):
        return [cmoobj.to_python_object() for cmoobj in self.body]
    def get_xml_body_tag(self):
        r=""
        for cmoobj in self.body:
            r=r+cmoobj.to_xml()
        return r
 
###################

class CError2(AbstractCMOWithCMO):
    def __init__(self,cmobj):
        AbstractCMOWithCMO.__init__(self,OpenXMConst.CMO_ERROR2,OpenXMConst.CMO_ERROR2_xmltagname,cmobj)


class CNull(AbstractCMO):
    def __init__(self):
        AbstractCMO.__init__(self,OpenXMConst.CMO_NULL,OpenXMConst.CMO_NULL_xmltagname,None)

    @classmethod
    def new_from_stream(self,stream,byte_order):
        return self()


class CInt32(AbstractCMO):
    def __init__(self,n):
        AbstractCMO.__init__(self,OpenXMConst.CMO_INT32,OpenXMConst.CMO_INT32_xmltagname,n)
    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        return self(n)

class CDatum(AbstractCMOWithBytes):
    def __init__(self,bytedata):
        AbstractCMOWithBytes.__init__(self,OpenXMConst.CMO_DATUM,OpenXMConst.CMO_DATUM_xmltagname,bytedata)


class CString(AbstractCMOWithBytes):
    def __init__(self,bytedata):
        AbstractCMOWithBytes.__init__(self,OpenXMConst.CMO_STRING,OpenXMConst.CMO_STRING_xmltagname,bytedata)
    def get_xml_sequence_tag(self):
        return '<string>"%s"</string>' % self.get_sequence()

class CMathcap(AbstractCMOWithCMO):
    def __init__(self,cmo_list):
        AbstractCMOWithCMO.__init__(self,OpenXMConst.CMO_MATHCAP,OpenXMConst.CMO_MATHCAP_xmltagname,cmo_list)
        l=cmo_list.to_python_object()
        self.system_info=l[0]
        self.available_sm_tag=l[1]
        self.available_ox_data_tags=l[2]
    def to_python_object(self):
        return tuple(self.body.to_python_object())

class CList(AbstractCMOWithSequence):
    def __init__(self,listofcmobj):
        AbstractCMOWithSequence.__init__(self,OpenXMConst.CMO_LIST,OpenXMConst.CMO_LIST_xmltagname,listofcmobj)

###################################################

class CZero(AbstractCMO):
    def __init__(self):
        AbstractCMO.__init__(self,OpenXMConst.CMO_ZERO,OpenXMConst.CMO_ZERO_xmltagname,None)
    def to_python_object(self):
        return 0

    @classmethod
    def new_from_stream(self,stream,byte_order):
        return self()


class CZZ(AbstractCMO):
    DIGIT=(2L)**32
    def __init__(self,n):
        AbstractCMO.__init__(self,OpenXMConst.CMO_ZZ,OpenXMConst.CMO_ZZ_xmltagname,n)
    def body_to_byte(self,byte_order):
        (f,b)=self.list_of_int32_from_long(self.body)
        r=struct.pack(byte_order+'i',f)
        for bi in b:
            r=r+ struct.pack(byte_order+'i',bi)
        return r

    def get_xml_body_tag(self):
        (f,b)=self.list_of_int32_from_long(self.body)
        r="<int32>%d</int32>" % f
        r=r+"".join("<int32>%d</int32>" % i for i in b)
        return r

    @classmethod
    def new_from_stream(self,stream,byte_order):
        f=cmo.read_int32_from_stream(stream,byte_order)
        b=[cmo.read_int32_from_stream(stream,byte_order) for i in range(abs(f))]
        l=self.long_from_list_of_int32(f,b)
        return self(l)


    @classmethod
    def long_from_list_of_int32(self,f,listofint32):
        r=0L
        n=abs(f)
        for i in range(n):
            r=r*self.DIGIT+listofint32[n-1-i]
        if f>0:
            return r
        else:
            return n
    @classmethod
    def list_of_int32_from_long(self,num):
        if num>0:
            sig=1
        else:
            sig=-1
        l=[]
        n=abs(num)
        while (n>0):
            r=n%self.DIGIT
            n= (n-r) / self.DIGIT
            l.append(r)
        return (sig*len(l),l)



class CQQ(AbstractCMOWithCMOTuple):
    def __init__(self,num,den):
         AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_QQ,OpenXMConst.CMO_QQ_xmltagname,(num,den))

    def body_to_byte(self,byte_order):
        (num,den)=self.body
        (f,a)=CZZ.list_of_int32_from_long(num)
        (n,b)=CZZ.list_of_int32_from_long(den)
        r=struct.pack(byte_order+'i',f)
        for ai in a:
            r=r+ struct.pack(byte_order+'i',ai)
        r=r+struct.pack(byte_order+'i',n)
        for bi in b:
            r=r+ struct.pack(byte_order+'i',bi)
        return r
    def get_xml_body_tag(self):
        (num,den)=self.body
        a=CZZ(num)
        b=CZZ(den)
        return a.get_xml_body_tag()+b.get_xml_body_tag()

    @classmethod
    def new_from_stream(self,stream,byte_order):
        f=cmo.read_int32_from_stream(stream,byte_order)
        b=[cmo.read_int32_from_stream(stream,byte_order) for i in range(abs(f))]
        num=CZZ.long_from_list_of_int32(f,b)

        f=cmo.read_int32_from_stream(stream,byte_order)
        b=[cmo.read_int32_from_stream(stream,byte_order) for i in range(abs(f))]
        den=CZZ.long_from_list_of_int32(f,b)
        return self(num,den)

class CRational(AbstractCMOWithCMOTuple):
    def __init__(self,num_cmo,den_cmo):
         AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_RATIONAL,OpenXMConst.CMO_RATIONAL_xmltagname,(num_cmo,den_cmo))
        
    @classmethod
    def new_from_stream(self,stream,byte_order):
        num=cmo.new_from_stream(stream,byte_order)
        den=cmo.new_from_stream(stream,byte_order)
        return self(num,den)


class CIndeterminate(AbstractCMOWithCMO):
    def __init__(self,label_cstring):
        AbstractCMOWithCMO.__init__(self,OpenXMConst.CMO_INDETERMINATE,OpenXMConst.CMO_INDETERMINATE_xmltagname,label_cstring)

class CTree(AbstractCMOWithCMOTuple):
    def __init__(self,name_cstring,attribute_cmolist,children_cmolist):
        AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_TREE,OpenXMConst.CMO_TREE_xmltagname,(name_cstring,attribute_cmolist,children_cmolist))
    @classmethod
    def new_from_stream(self,stream,byte_order):
        r1=cmo.new_from_stream(stream,byte_order)
        r2=cmo.new_from_stream(stream,byte_order)
        r3=cmo.new_from_stream(stream,byte_order)
        return self(r1,r2,r3)


class CLambda(AbstractCMOWithCMOTuple):
    def __init__(self,arg_list,body_tree):
        AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_LAMBDA,OpenXMConst.CMO_LAMBDA_xmltagname,(arg_list,body_tree))
    @classmethod
    def new_from_stream(self,stream,byte_order):
        r1=cmo.new_from_stream(stream,byte_order)
        r2=cmo.new_from_stream(stream,byte_order)
        return self(r1,r2)

class CMonomial32(AbstractCMO):
    def __init__(self,index_list_int,coef_cmoobj):
        AbstractCMO.__init__(self,OpenXMConst.CMO_MONOMIAL32,OpenXMConst.CMO_MONOMIAL32_xmltagname,(index_list_int,coef_cmoobj))

    def body_to_byte(self,byte_order):
        (ind,coef)=self.body
        form=byte_order+'i'
        r=struct.pack(form,len(ind))
        for ai in ind:
            r=r+struct.pack(form,ai)
        return r+coef.to_byte(byte_order)
    def to_python_object(self):
        (ind,coef)=self.body
        return (ind,coef.to_python_object()) 
    def get_xml_body_tag(self):
        (ind,coef)=self.body
        r= "<int32>%d</int32>" % len(ind)
        for ai in ind:
            r= "<int32>%d</int32>" % ai
        return r+coef.to_xml()
    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        ind=[cmo.read_int32_from_stream(stream,byte_order) for i in range(n)]
        coef=cmo.new_from_stream(stream,byte_order)
        return self(ind,coef)

class CDMSGeneric(AbstractCMO):
    def __init__(self):
        AbstractCMO.__init__(self,OpenXMConst.CMO_DMS_GENERIC,OpenXMConst.CMO_DMS_GENERIC_xmltagname,None)
    def to_python_object(self):
        return None

    @classmethod
    def new_from_stream(self,stream,byte_order):
        return self()

class CRingByName(AbstractCMOWithCMO):
    def __init__(self,name_cstring):
        AbstractCMOWithCMO.__init__(self,OpenXMConst.CMO_RING_BY_NAME,OpenXMConst.CMO_RING_BY_NAME_xmltagname,name_cstring)

class CDMSOfNVariables(AbstractCMOWithCMO):
    def __init__(self,definition_cmolist):
        AbstractCMOWithCMO.__init__(self,OpenXMConst.CMO_DMS_OF_N_VARIABLES,OpenXMConst.CMO_DMS_OF_N_VARIABLES_xmltagname,definition_cmolist)
    def to_python_object(self):
        b=self.body.to_python_object()
        n_variables=b[0]
        p=b[1]
        if len(b)>2 :
            additionalinfo=tuple(b[2:])
        else:
            additionalinfo=None
        return (n_variables,p,additionalinfo)

class CDistributedPolynomial(AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence):
    def __init__(self,dring_definition_cmo,monomials_list_of_cmoobj):
        AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence.__init__(self,OpenXMConst.CMO_DISTRIBUTED_POLYNOMIAL,OpenXMConst.CMO_DISTRIBUTED_POLYNOMIAL_xmltagname,(dring_definition_cmo,),monomials_list_of_cmoobj)
    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        dring_definition_cmo=cmo.new_from_stream(stream,byte_order)
        listofcmobj=[cmo.new_from_stream(stream,byte_order) for i in range(n)]
        return self(dring_definition_cmo,listofcmobj)



class CPolynomialInOneVariable(AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence):
    def __init__(self,name_of_var_int,expandcoef_list_of_tuple):
        AbstractCMOWithFixedLengthInfomationAndVariableLengthSequence.__init__(self,OpenXMConst.CMO_POLYNOMIAL_IN_ONE_VARIABLE,OpenXMConst.CMO_POLYNOMIAL_IN_ONE_VARIABLE_xmltagname,(name_of_var_int,),expandcoef_list_of_tuple)

    def sequence_to_byte(self,byte_order):
        r=''
        for (e,cmobj) in self.get_sequence():
            r=r+struct.pack(byte_order+'i',e) 
            r=r+cmobj.to_byte(byte_order)
        return r
    def info_to_byte(self,byte_order):
        info=self.get_info()
        r=''
        for a in info:
            r=r+struct.pack(byte_order+'i',a)
        return r
    def body_to_byte(self,byte_order):
        return struct.pack(byte_order+'i',self.get_len_of_sequence())+self.info_to_byte(byte_order)+self.sequence_to_byte(byte_order)


    def sequence_to_python_object(self):
        return [(e,coef.to_python_object()) for (e,coef) in self.get_sequence()]
    def info_to_python_object(self):
        return self.get_info()

    def get_xml_sequence_tag(self):
        return "".join(('<int32>%d</int32>' % e)+coef.to_xml() for (e,coef) in self.get_sequence())
    def get_xml_info_tag(self):
        info=self.get_info()[0]
        return '<int32>%d</int32>' % info


    @classmethod
    def new_from_stream(self,stream,byte_order):
        n=cmo.read_int32_from_stream(stream,byte_order)
        name_of_var=cmo.read_int32_from_stream(stream,byte_order)
        expandcoef_list_of_tuple=[]
        for i in  range(n):
            e=cmo.read_int32_from_stream(stream,byte_order)
            coef=cmo.new_from_stream(stream,byte_order)
            expandcoef_list_of_tuple.append( (e,coef) )
        return self(name_of_var,expandcoef_list_of_tuple)

        

class CRecursivePolynomial(AbstractCMOWithCMOTuple):
    def __init__(self,var_clist,polynomialorcoef_cmo):
         AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_RECURSIVE_POLYNOMIAL,OpenXMConst.CMO_RECURSIVE_POLYNOMIAL_xmltagname,(var_clist,polynomialorcoef_cmo))
        
    @classmethod
    def new_from_stream(self,stream,byte_order):
        var=cmo.new_from_stream(stream,byte_order)
        poly=cmo.new_from_stream(stream,byte_order)
        return self(var,poly)

class CBigFloat(AbstractCMOWithCMOTuple):
    def __init__(self,a_ZZ,e_ZZ):
         AbstractCMOWithCMOTuple.__init__(self,OpenXMConst.CMO_BIGFLOAT,OpenXMConst.CMO_BIGFLOAT_xmltagname,(a_ZZ,e_ZZ))

    def to_python_object(self):
        (a_ZZ,e_ZZ)=self.body
        a=a.to_python_object()
        e=e.to_python_object()
        return a*((2L)**(a))
    @classmethod
    def new_from_stream(self,stream,byte_order):
        a=cmo.new_from_stream(stream,byte_order)
        e=cmo.new_from_stream(stream,byte_order)
        return self(a,e)

            
class CIEEEDoubleFloat(AbstractCMO):
    def __init__(self,d):
        AbstractCMO.__init__(self,OpenXMConst.CMO_IEEE_DOUBLE_FLOAT,OpenXMConst.CMO_IEEE_DOUBLE_FLOAT_xmltagname,d)

    def body_to_byte(self,byte_order):
        return struct.pack(byte_order+'d',self.body)

    def get_xml_body_tag(self):
        return "".join("<byte>%d</byte>" % i for i in self.body_to_byte('!'))


    @classmethod
    def new_from_stream(self,stream,byte_order):
        d=cmo.read_ieeefoat64_from_stream(stream,byte_order)
        return self(d)



class cmo:
    implemented={
        OpenXMConst.CMO_NULL:
            CNull,
        OpenXMConst.CMO_INT32:
            CInt32,
        OpenXMConst.CMO_ERROR2:
            CError2,
        OpenXMConst.CMO_MATHCAP:
            CMathcap,
        OpenXMConst.CMO_DATUM:
            CDatum,
        OpenXMConst.CMO_STRING:
            CString,
        OpenXMConst.CMO_LIST:
            CList,


        OpenXMConst.CMO_ZERO:
            CZero,
        OpenXMConst.CMO_ZZ:
            CZZ,
        OpenXMConst.CMO_QQ:
            CQQ,
        OpenXMConst.CMO_INDETERMINATE:
            CIndeterminate,
        OpenXMConst.CMO_TREE:
            CTree,
        OpenXMConst.CMO_LAMBDA:
            CLambda,
        OpenXMConst.CMO_MONOMIAL32:
            CMonomial32,
        OpenXMConst.CMO_DMS_GENERIC:
            CDMSGeneric,
        OpenXMConst.CMO_RING_BY_NAME:
            CRingByName,
        OpenXMConst.CMO_DMS_OF_N_VARIABLES:
            CDMSOfNVariables,
        OpenXMConst.CMO_DISTRIBUTED_POLYNOMIAL:
            CDistributedPolynomial,
        OpenXMConst.CMO_POLYNOMIAL_IN_ONE_VARIABLE:
            CPolynomialInOneVariable,
        OpenXMConst.CMO_RECURSIVE_POLYNOMIAL:
            CRecursivePolynomial,
        OpenXMConst.CMO_BIGFLOAT:
            CBigFloat,
        OpenXMConst.CMO_IEEE_DOUBLE_FLOAT:
            CIEEEDoubleFloat,
        }

# not implemented
#    CMO_LIST_R = 28
#    CMO_INT32COEFF = 30
#
#    CMO_64BIT_MACHINE_DOUBLE = 40
#    CMO_ARRAY_OF_64BIT_MACHINE_DOUBLE = 41
#    CMO_128BIT_MACHINE_DOUBLE = 42
#    CMO_ARRAY_OF_128BIT_MACHINE_DOUBLE = 43
    


    @classmethod
    def read_ieeefoat64_from_stream(self,stream,byte_order):
        return struct.unpack(byte_order+'d',stream.read(8))[0]

    @classmethod
    def read_int32_from_stream(self,stream,byte_order):
        return struct.unpack(byte_order+'i',stream.read(4))[0]

    @classmethod
    def get_implemented_cmo(self):
        return [(OpenXMConst.OX_DATA,self.implemented.keys())]
    @classmethod
    def new_from_stream(self,stream,byte_order):
        tag=cmo.read_int32_from_stream(stream,byte_order)
        if tag in self.implemented:
            return self.implemented[tag].new_from_stream(stream,byte_order)
        else:
            return None
#        if tag==OpenXMConst.CMO_NULL:
#            return Null.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_INT32:
#            return Int32.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_ERROR2:
#            return Error2.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_MATHCAP:
#            return Mathcap.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_DATUM:
#            return Datum.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_STRING:
#            return CString.new_from_stream(stream,byte_order)
#        elif tag==OpenXMConst.CMO_LIST:
#            return List.new_from_stream(stream,byte_order)



def test_main():
    print "/* Classes for the following CMO Objects are implemented. */"
    for name in dir(OpenXMConst):
        if name.startswith("CMO_"):
            num=OpenXMConst.__dict__[name]
            if num in cmo.implemented:
                print "#define ",name,"\t",num


if __name__ == "__main__":
    test_main()


